# The following function loads the requested manifold using the
# following conventions:  
#
#   1. Numbers in parens at the end mean do Dehn filling on the loaded
#   manifold, e.g. m125(1,2)(4,5) means do (1,2) filling on the first
#   cusp and (4,5) filling on the second cusp.
#
#   2. Names of the form m123, s123, v123, and so on mean to load the
#   corresponding Census manifold.
#
#   3. Names of the form 4_1, 04_1, 4_01, 5^2_6, 6_4^7, etc, mean load the
#   complement of the link with that name.  Similary L20935.  l104001, etc.
#
#   4. Names of the form b++LLR, b+-llR, bo-RRL, bn+LRLR load the
#   correponding torus bundle.
#
#  If one of the above rules does _not_ apply, it looks for the named
#  manifold in the current directory and looks in the path given by
#  the user variable SNAPPEA_MANIFOLD_DIRECTORY, which and checks the
#  default path defined in paths.py.
#
#  The directories for loading the Census manifolds and the knot
#  manifolds (assumed to be in link file format), as well as the default
#  directory to look for manifolds are contained in the file paths.py

from SnapPea import *
from function_wrapper_modifications import *

import os, sys, re, string, paths, tempfile

# for speed we compile our regular expressions

split_filling_info = re.compile("(.*?)((?:\([0-9 .+-]+,[0-9 .+-]+\))+)")
is_census_manifold = re.compile("([msv])(\d+)$")
is_torus_bundle = re.compile("b([+-no])([+-])([lLrR]+)$")
is_knot_complement = re.compile("(\d+)_(\d+)$")
is_link_complement1 = re.compile("(?P<crossings>\d+)[\^](?P<components>\d+)[_](?P<index>\d+)$")
is_link_complement2 = re.compile("(?P<crossings>\d+)[_](?P<index>\d+)[\^](?P<components>\d+)$")
is_link_complement3 = re.compile("[lL](?P<components>\d)(?P<crossings>\d\d)(?P<index>\d\d\d)$")


def get_manifold(name):
    # get filling info, if any
    m = split_filling_info.match(name)
    if m:
        real_name = m.group(1)
        fillings = re.subn("\)\(", "),(", m.group(2))[0]
        fillings = eval( "[" + fillings + "]" )
    else:
        real_name = name
        fillings = ()

    # Check if name is one of the special types of manifold.
    
    while 1:
        m = is_census_manifold.match(real_name)
        if m:
            if m.group(1) == 'm':
                census = 5
            elif m.group(1) == 's':
                census = 6
            else:
                census = 7
            manifold = GetCensusManifold(census, int(m.group(2)))
            break

        m = is_torus_bundle.match(real_name)
        if m:
            LRstring = string.upper(m.group(3))

            if m.group(1) == '+' or m.group(1) == 'o':
                negative_determinant = 0
            else:
                negative_determinant = 1

            if m.group(2) == '+':
                negative_trace = 0
            else:
                negative_trace = 1

            manifold =  easy_triangulate_punctured_torus_bundle(negative_determinant,
                                                                negative_trace, LRstring);
            break

        m = is_knot_complement.match(real_name)
        if m:
            try:
                manifold = get_triangulation(
                    paths.manifold_directory + os.sep + "L1%.2d%.3d" % (int(m.group(1)), int(m.group(2))))
                if not manifold: raise ""
                set_triangulation_name(manifold, real_name)
                break
            except:
                raise IOError, "Requested link complement " + real_name + " not found."

        m = is_link_complement1.match(real_name)
        if not m:
            m = is_link_complement2.match(real_name)
        if not m:
            m = is_link_complement3.match(real_name)
        if m:
            try:
                manifold = get_triangulation(
                    paths.manifold_directory + os.sep + "L%.1d%.2d%.3d" % (int(m.group("components")),
                                                        int(m.group("crossings")), int(m.group("index"))))
                if not manifold: raise ""
                set_triangulation_name(manifold, real_name)
                break
            except:
                raise IOError, "Requested link complement " + real_name + " not found."

        # Try to load manifold from file.
        locations = [os.curdir, os.environ["SNAPPEA_MANIFOLD_DIRECTORY"],
                     paths.manifold_directory]
        found = 0
        for location in locations:
            try:
                f = open(location + os.sep + real_name)
            except IOError:
                continue
            f.close()
            found = 1
            break

        if found:
            manifold = get_triangulation(location + os.sep + real_name)
            break
        else:
            # We've failed to locate the manifold
            raise IOError, "Requested manifold %s not found." % real_name
        
    # do the dehn filling

    if len(fillings) > 0:
        if len(fillings) > get_num_cusps(manifold):
            raise ValueError, "More fillings requested than manifold has cusps"
        for i in range(len(fillings)):
            m, l = fillings[i]
            set_cusp_info(manifold, i, 0, m, l)
        do_Dehn_filling(manifold)
        s = real_name
        for filling in fillings:
            s = s + "(%s,%s)" % filling
        set_triangulation_name(manifold, s)

    manifold.thisown = 1
    return manifold

#end of get_manifold

def gluing_equations(manifold):
    fname = tempfile.mktemp()
    if std_complex_equations_to_file(manifold, fname):
        f = open(fname)
        edge = []
        cusp = []
        for line in f.readlines():
            if line[:4] == "edge":
                curr = edge
            elif line[:4] == "cusp":
                curr = cusp
            else:
                curr.append( map(int, string.split(line)))
    else:
        raise RuntimeError, "Gluing equation failure"

    os.remove(fname)
    return [edge, cusp]
            
    

#For backwards compatibility:

get_named_manifold = get_manifold

def are_isometric(M, N):
    ans, p1, p2 = compute_isometries(M, N)
    if p1 != "NULL":
        free_isometry_list(p1)
    if p2 != "NULL":
        free_isometry_list(p2)

    return ans

def is_two_bridge(manifold):
    temp_mfld = copy_triangulation(manifold)
    canonize(temp_mfld)
    return two_bridge(temp_mfld)  












